@@ -1,22 +1,26 @@
-from __future__ import absolute_import, division, print_function
-
 import functools
+from collections import OrderedDict
 
 import numpy as np
 
 from .. import Variable
 from ..core import indexing
-from ..core.pycompat import OrderedDict, bytes_type, iteritems, unicode_type
 from ..core.utils import FrozenOrderedDict, close_on_error
 from .common import WritableCFDataStore
 from .file_manager import CachingFileManager
 from .locks import HDF5_LOCK, combine_locks, ensure_lock, get_write_lock
 from .netCDF4_ import (
-    BaseNetCDF4Array, GroupWrapper, _encode_nc4_variable,
-    _extract_nc4_variable_encoding, _get_datatype, _nc4_require_group)
+    BaseNetCDF4Array, _encode_nc4_variable, _extract_nc4_variable_encoding,
+    _get_datatype, _nc4_require_group)
 
 
 class H5NetCDFArrayWrapper(BaseNetCDF4Array):
+
+    def get_array(self, needs_lock=True):
+        ds = self.datastore._acquire(needs_lock)
+        variable = ds.variables[self.variable_name]
+        return variable
+
     def __getitem__(self, key):
         return indexing.explicit_indexing_adapter(
             key, self.shape, indexing.IndexingSupport.OUTER_1VECTOR,
@@ -26,13 +30,13 @@ class H5NetCDFArrayWrapper(BaseNetCDF4Array):
         # h5py requires using lists for fancy indexing:
         # https://github.com/h5py/h5py/issues/992
         key = tuple(list(k) if isinstance(k, np.ndarray) else k for k in key)
-        array = self.get_array()
         with self.datastore.lock:
+            array = self.get_array(needs_lock=False)
             return array[key]
 
 
 def maybe_decode_bytes(txt):
-    if isinstance(txt, bytes_type):
+    if isinstance(txt, bytes):
         return txt.decode('utf-8')
     else:
         return txt
@@ -59,26 +63,19 @@ def _h5netcdf_create_group(dataset, name):
     return dataset.create_group(name)
 
 
-def _open_h5netcdf_group(filename, mode, group):
-    import h5netcdf
-    ds = h5netcdf.File(filename, mode=mode)
-    with close_on_error(ds):
-        ds = _nc4_require_group(
-            ds, group, mode, create_group=_h5netcdf_create_group)
-    return GroupWrapper(ds)
-
-
 class H5NetCDFStore(WritableCFDataStore):
     """Store for reading and writing data via h5netcdf
     """
 
     def __init__(self, filename, mode='r', format=None, group=None,
                  lock=None, autoclose=False):
+        import h5netcdf
+
         if format not in [None, 'NETCDF4']:
             raise ValueError('invalid format for h5netcdf backend')
+
         self._manager = CachingFileManager(
-            _open_h5netcdf_group, filename, mode=mode,
-            kwargs=dict(group=group))
+            h5netcdf.File, filename, mode=mode)
 
         if lock is None:
             if mode == 'r':
@@ -86,15 +83,22 @@ class H5NetCDFStore(WritableCFDataStore):
             else:
                 lock = combine_locks([HDF5_LOCK, get_write_lock(filename)])
 
+        self._group = group
         self.format = format
         self._filename = filename
         self._mode = mode
         self.lock = ensure_lock(lock)
         self.autoclose = autoclose
 
+    def _acquire(self, needs_lock=True):
+        with self._manager.acquire_context(needs_lock) as root:
+            ds = _nc4_require_group(root, self._group, self._mode,
+                                    create_group=_h5netcdf_create_group)
+        return ds
+
     @property
     def ds(self):
-        return self._manager.acquire().value
+        return self._acquire()
 
     def open_store_variable(self, name, var):
         import h5py
@@ -124,7 +128,7 @@ class H5NetCDFStore(WritableCFDataStore):
         encoding['original_shape'] = var.shape
 
         vlen_dtype = h5py.check_dtype(vlen=var.dtype)
-        if vlen_dtype is unicode_type:
+        if vlen_dtype is str:
             encoding['dtype'] = str
         elif vlen_dtype is not None:  # pragma: no cover
             # xarray doesn't support writing arbitrary vlen dtypes yet.
@@ -136,7 +140,7 @@ class H5NetCDFStore(WritableCFDataStore):
 
     def get_variables(self):
         return FrozenOrderedDict((k, self.open_store_variable(k, v))
-                                 for k, v in iteritems(self.ds.variables))
+                                 for k, v in self.ds.variables.items())
 
     def get_attrs(self):
         return FrozenOrderedDict(_read_attributes(self.ds))
@@ -182,7 +186,7 @@ class H5NetCDFStore(WritableCFDataStore):
                 'NC_CHAR type.' % name)
 
         if dtype is str:
-            dtype = h5py.special_dtype(vlen=unicode_type)
+            dtype = h5py.special_dtype(vlen=str)
 
         encoding = _extract_h5nc_encoding(variable,
                                           raise_on_invalid=check_encoding)
@@ -208,10 +212,12 @@ class H5NetCDFStore(WritableCFDataStore):
 
         encoding['chunks'] = encoding.pop('chunksizes', None)
 
-        for key in ['compression', 'compression_opts', 'shuffle',
-                    'chunks', 'fletcher32']:
-            if key in encoding:
-                kwargs[key] = encoding[key]
+        # Do not apply compression, filters or chunking to scalars.
+        if variable.shape:
+            for key in ['compression', 'compression_opts', 'shuffle',
+                        'chunks', 'fletcher32']:
+                if key in encoding:
+                    kwargs[key] = encoding[key]
         if name not in self.ds:
             nc4_var = self.ds.create_variable(
                 name, dtype=dtype, dimensions=variable.dims,
@@ -219,7 +225,7 @@ class H5NetCDFStore(WritableCFDataStore):
         else:
             nc4_var = self.ds[name]
 
-        for k, v in iteritems(attrs):
+        for k, v in attrs.items():
             nc4_var.attrs[k] = v
 
         target = H5NetCDFArrayWrapper(name, self)
@@ -228,9 +234,6 @@ class H5NetCDFStore(WritableCFDataStore):
 
     def sync(self):
         self.ds.sync()
-        # if self.autoclose:
-        #     self.close()
-        # super(H5NetCDFStore, self).sync(compute=compute)
 
     def close(self, **kwargs):
         self._manager.close(**kwargs)
