@@ -1,23 +1,17 @@
-from __future__ import absolute_import, division, print_function
-
 import functools
 import operator
 from collections import defaultdict
+from contextlib import suppress
 from datetime import timedelta
+from typing import Sequence
 
 import numpy as np
 import pandas as pd
 
 from . import duck_array_ops, nputils, utils
-from .pycompat import (
-    dask_array_type, integer_types, iteritems, range, suppress)
+from .pycompat import dask_array_type, integer_types
 from .utils import is_dict_like
 
-try:
-    from collections.abc import Hashable
-except ImportError:  # Py2
-    from collections import Hashable
-
 
 def expanded_indexer(key, ndim):
     """Given a key for indexing an ndarray, return an equivalent key which is a
@@ -158,7 +152,7 @@ def convert_label_indexer(index, label, index_name='', method=None,
         else:
             for k, v in label.items():
                 # index should be an item (i.e. Hashable) not an array-like
-                if not isinstance(v, Hashable):
+                if isinstance(v, Sequence) and not isinstance(v, str):
                     raise ValueError('Vectorized selection is not '
                                      'available along level variable: ' + k)
             indexer, new_index = index.get_loc_level(
@@ -214,7 +208,7 @@ def get_dim_indexers(data_obj, indexers):
 
     level_indexers = defaultdict(dict)
     dim_indexers = {}
-    for key, label in iteritems(indexers):
+    for key, label in indexers.items():
         dim, = data_obj[key].dims
         if key != dim:
             # assume here multi-index level indexer
@@ -222,7 +216,7 @@ def get_dim_indexers(data_obj, indexers):
         else:
             dim_indexers[key] = label
 
-    for dim, level_labels in iteritems(level_indexers):
+    for dim, level_labels in level_indexers.items():
         if dim_indexers.get(dim, False):
             raise ValueError("cannot combine multi-index level indexers "
                              "with an indexer for dimension %s" % dim)
@@ -243,7 +237,7 @@ def remap_label_indexers(data_obj, indexers, method=None, tolerance=None):
     new_indexes = {}
 
     dim_indexers = get_dim_indexers(data_obj, indexers)
-    for dim, label in iteritems(dim_indexers):
+    for dim, label in dim_indexers.items():
         try:
             index = data_obj.indexes[dim]
         except KeyError:
@@ -300,7 +294,7 @@ def _index_indexer_1d(old_indexer, applied_indexer, size):
     return indexer
 
 
-class ExplicitIndexer(object):
+class ExplicitIndexer:
     """Base class for explicit indexer objects.
 
     ExplicitIndexer objects wrap a tuple of values given by their ``tuple``
@@ -358,7 +352,7 @@ class BasicIndexer(ExplicitIndexer):
                                 .format(type(self).__name__, k))
             new_key.append(k)
 
-        super(BasicIndexer, self).__init__(new_key)
+        super().__init__(new_key)
 
 
 class OuterIndexer(ExplicitIndexer):
@@ -394,7 +388,7 @@ class OuterIndexer(ExplicitIndexer):
                                 .format(type(self).__name__, k))
             new_key.append(k)
 
-        super(OuterIndexer, self).__init__(new_key)
+        super().__init__(new_key)
 
 
 class VectorizedIndexer(ExplicitIndexer):
@@ -433,10 +427,10 @@ class VectorizedIndexer(ExplicitIndexer):
                                 .format(type(self).__name__, k))
             new_key.append(k)
 
-        super(VectorizedIndexer, self).__init__(new_key)
+        super().__init__(new_key)
 
 
-class ExplicitlyIndexed(object):
+class ExplicitlyIndexed:
     """Mixin to mark support for Indexer subclasses in indexing."""
 
 
@@ -459,7 +453,13 @@ class ImplicitToExplicitIndexingAdapter(utils.NDArrayMixin):
 
     def __getitem__(self, key):
         key = expanded_indexer(key, self.ndim)
-        return self.array[self.indexer_cls(key)]
+        result = self.array[self.indexer_cls(key)]
+        if isinstance(result, ExplicitlyIndexed):
+            return type(self)(result, self.indexer_cls)
+        else:
+            # Sometimes explicitly indexed arrays return NumPy arrays or
+            # scalars.
+            return result
 
 
 class LazilyOuterIndexedArray(ExplicitlyIndexedNDArrayMixin):
@@ -746,7 +746,7 @@ def _combine_indexers(old_key, shape, new_key):
                                    np.broadcast_arrays(*old_key.tuple)))
 
 
-class IndexingSupport(object):  # could inherit from enum.Enum on Python 3
+class IndexingSupport:  # could inherit from enum.Enum on Python 3
     # for backends that support only basic indexer
     BASIC = 'BASIC'
     # for backends that support basic / outer indexer
