@@ -1,13 +1,22 @@
-from __future__ import absolute_import, division, print_function
+from collections import OrderedDict
+from typing import (Any, Dict, Hashable, List, Mapping, Optional, Sequence,
+                    Set, Tuple, Union)
 
 import pandas as pd
 
+from . import dtypes, pdcompat
 from .alignment import deep_align
-from .pycompat import OrderedDict, basestring
+from .pycompat import TYPE_CHECKING
 from .utils import Frozen
-from .variable import as_variable, assert_unique_multiindex_level_names
+from .variable import (
+    Variable, as_variable, assert_unique_multiindex_level_names)
 
-PANDAS_TYPES = (pd.Series, pd.DataFrame, pd.Panel)
+if TYPE_CHECKING:
+    from .dataarray import DataArray
+    from .dataset import Dataset
+
+
+PANDAS_TYPES = (pd.Series, pd.DataFrame, pdcompat.Panel)
 
 _VALID_COMPAT = Frozen({'identical': 0,
                         'equals': 1,
@@ -16,13 +25,14 @@ _VALID_COMPAT = Frozen({'identical': 0,
                         'no_conflicts': 4})
 
 
-def broadcast_dimension_size(variables):
-    # type: (List[Variable],) -> Variable
+def broadcast_dimension_size(
+    variables: List[Variable],
+) -> 'OrderedDict[Any, int]':
     """Extract dimension sizes from a dictionary of variables.
 
     Raises ValueError if any dimensions have different sizes.
     """
-    dims = OrderedDict()
+    dims = OrderedDict()  # type: OrderedDict[Any, int]
     for var in variables:
         for dim, size in zip(var.dims, var.shape):
             if dim in dims and size != dims[dim]:
@@ -99,7 +109,7 @@ class OrderedDefaultDict(OrderedDict):
     # beware: does not pickle or copy properly
     def __init__(self, default_factory):
         self.default_factory = default_factory
-        super(OrderedDefaultDict, self).__init__()
+        super().__init__()
 
     def __missing__(self, key):
         self[key] = default = self.default_factory()
@@ -143,15 +153,15 @@ def merge_variables(
 
     # n.b. it's important to fill up merged in the original order in which
     # variables appear
-    merged = OrderedDict()
+    merged = OrderedDict()  # type: OrderedDict[Any, Variable]
 
-    for name, variables in lookup.items():
+    for name, var_list in lookup.items():
         if name in priority_vars:
             # one of these arguments (e.g., the first for in-place arithmetic
             # or the second for Dataset.update) takes priority
             merged[name] = priority_vars[name]
         else:
-            dim_variables = [var for var in variables if (name,) == var.dims]
+            dim_variables = [var for var in var_list if (name,) == var.dims]
             if dim_variables:
                 # if there are dimension coordinates, these must be equal (or
                 # identical), and they take priority over non-dimension
@@ -159,7 +169,7 @@ def merge_variables(
                 merged[name] = unique_variable(name, dim_variables, dim_compat)
             else:
                 try:
-                    merged[name] = unique_variable(name, variables, compat)
+                    merged[name] = unique_variable(name, var_list, compat)
                 except MergeError:
                     if compat != 'minimal':
                         # we need more than "minimal" compatibility (for which
@@ -169,8 +179,9 @@ def merge_variables(
     return merged
 
 
-def expand_variable_dicts(list_of_variable_dicts):
-    # type: (List[Union[Dataset, Dict]]) -> List[Dict[Any, Variable]]
+def expand_variable_dicts(
+    list_of_variable_dicts: 'List[Union[Dataset, OrderedDict]]',
+) -> 'List[Mapping[Any, Variable]]':
     """Given a list of dicts with xarray object values, expand the values.
 
     Parameters
@@ -188,29 +199,30 @@ def expand_variable_dicts(list_of_variable_dicts):
     an input's values. The values of each ordered dictionary are all
     xarray.Variable objects.
     """
-    from .dataarray import DataArray
+    from .dataarray import DataArray  # noqa: F811
     from .dataset import Dataset
 
     var_dicts = []
 
     for variables in list_of_variable_dicts:
         if isinstance(variables, Dataset):
-            sanitized_vars = variables.variables
-        else:
-            # append coords to var_dicts before appending sanitized_vars,
-            # because we want coords to appear first
-            sanitized_vars = OrderedDict()
+            var_dicts.append(variables.variables)
+            continue
 
-            for name, var in variables.items():
-                if isinstance(var, DataArray):
-                    # use private API for speed
-                    coords = var._coords.copy()
-                    # explicitly overwritten variables should take precedence
-                    coords.pop(name, None)
-                    var_dicts.append(coords)
+        # append coords to var_dicts before appending sanitized_vars,
+        # because we want coords to appear first
+        sanitized_vars = OrderedDict()  # type: OrderedDict[Any, Variable]
 
-                var = as_variable(var, name=name)
-                sanitized_vars[name] = var
+        for name, var in variables.items():
+            if isinstance(var, DataArray):
+                # use private API for speed
+                coords = var._coords.copy()
+                # explicitly overwritten variables should take precedence
+                coords.pop(name, None)
+                var_dicts.append(coords)
+
+            var = as_variable(var, name=name)
+            sanitized_vars[name] = var
 
         var_dicts.append(sanitized_vars)
 
@@ -233,11 +245,11 @@ def determine_coords(list_of_variable_dicts):
         All variable found in the input should appear in either the set of
         coordinate or non-coordinate names.
     """
-    from .dataarray import DataArray
+    from .dataarray import DataArray  # noqa: F811
     from .dataset import Dataset
 
-    coord_names = set()
-    noncoord_names = set()
+    coord_names = set()  # type: set
+    noncoord_names = set()  # type: set
 
     for variables in list_of_variable_dicts:
         if isinstance(variables, Dataset):
@@ -268,8 +280,8 @@ def coerce_pandas_values(objects):
     List of Dataset or OrderedDict objects. Any inputs or values in the inputs
     that were pandas objects have been converted into native xarray objects.
     """
+    from .dataarray import DataArray  # noqa: F811
     from .dataset import Dataset
-    from .dataarray import DataArray
 
     out = []
     for obj in objects:
@@ -339,7 +351,7 @@ def expand_and_merge_variables(objs, priority_arg=None):
 
 
 def merge_coords(objs, compat='minimal', join='outer', priority_arg=None,
-                 indexes=None):
+                 indexes=None, fill_value=dtypes.NA):
     """Merge coordinate variables.
 
     See merge_core below for argument descriptions. This works similarly to
@@ -348,7 +360,8 @@ def merge_coords(objs, compat='minimal', join='outer', priority_arg=None,
     """
     _assert_compat_valid(compat)
     coerced = coerce_pandas_values(objs)
-    aligned = deep_align(coerced, join=join, copy=False, indexes=indexes)
+    aligned = deep_align(coerced, join=join, copy=False, indexes=indexes,
+                         fill_value=fill_value)
     expanded = expand_variable_dicts(aligned)
     priority_vars = _get_priority_vars(aligned, priority_arg, compat=compat)
     variables = merge_variables(expanded, priority_vars, compat=compat)
@@ -389,12 +402,17 @@ def assert_valid_explicit_coords(variables, dims, explicit_coords):
                 'by the xarray data model.' % coord_name)
 
 
-def merge_core(objs,
-               compat='broadcast_equals',
-               join='outer',
-               priority_arg=None,
-               explicit_coords=None,
-               indexes=None):
+def merge_core(
+    objs,
+    compat='broadcast_equals',
+    join='outer',
+    priority_arg=None,
+    explicit_coords=None,
+    indexes=None,
+    fill_value=dtypes.NA
+) -> Tuple['OrderedDict[Hashable, Variable]',
+           Set[Hashable],
+           Dict[Hashable, int]]:
     """Core logic for merging labeled objects.
 
     This is not public API.
@@ -413,6 +431,8 @@ def merge_core(objs,
         An explicit list of variables from `objs` that are coordinates.
     indexes : dict, optional
         Dictionary with values given by pandas.Index objects.
+    fill_value : scalar, optional
+        Value to use for newly missing values
 
     Returns
     -------
@@ -432,7 +452,8 @@ def merge_core(objs,
     _assert_compat_valid(compat)
 
     coerced = coerce_pandas_values(objs)
-    aligned = deep_align(coerced, join=join, copy=False, indexes=indexes)
+    aligned = deep_align(coerced, join=join, copy=False, indexes=indexes,
+                         fill_value=fill_value)
     expanded = expand_variable_dicts(aligned)
 
     coord_names, noncoord_names = determine_coords(coerced)
@@ -457,10 +478,10 @@ def merge_core(objs,
                          'coordinates or not in the merged result: %s'
                          % ambiguous_coords)
 
-    return variables, coord_names, dict(dims)
+    return variables, coord_names, dims
 
 
-def merge(objects, compat='no_conflicts', join='outer'):
+def merge(objects, compat='no_conflicts', join='outer', fill_value=dtypes.NA):
     """Merge any number of xarray objects into a single Dataset as variables.
 
     Parameters
@@ -482,6 +503,8 @@ def merge(objects, compat='no_conflicts', join='outer'):
           of all non-null values.
     join : {'outer', 'inner', 'left', 'right', 'exact'}, optional
         How to combine objects with different indexes.
+    fill_value : scalar, optional
+        Value to use for newly missing values
 
     Returns
     -------
@@ -512,27 +535,36 @@ def merge(objects, compat='no_conflicts', join='outer'):
     --------
     concat
     """  # noqa
-    from .dataarray import DataArray
+    from .dataarray import DataArray  # noqa: F811
     from .dataset import Dataset
 
-    dict_like_objects = [
-        obj.to_dataset() if isinstance(obj, DataArray) else obj
-        for obj in objects]
+    dict_like_objects = list()
+    for obj in objects:
+        if not (isinstance(obj, (DataArray, Dataset, dict))):
+            raise TypeError("objects must be an iterable containing only "
+                            "Dataset(s), DataArray(s), and dictionaries.")
 
-    variables, coord_names, dims = merge_core(dict_like_objects, compat, join)
-    merged = Dataset._construct_direct(variables, coord_names, dims)
+        obj = obj.to_dataset() if isinstance(obj, DataArray) else obj
+        dict_like_objects.append(obj)
+
+    variables, coord_names, dims = merge_core(dict_like_objects, compat, join,
+                                              fill_value=fill_value)
+    # TODO: don't always recompute indexes
+    merged = Dataset._construct_direct(
+        variables, coord_names, dims, indexes=None)
 
     return merged
 
 
-def dataset_merge_method(dataset, other, overwrite_vars, compat, join):
+def dataset_merge_method(dataset, other, overwrite_vars, compat, join,
+                         fill_value=dtypes.NA):
     """Guts of the Dataset.merge method."""
 
     # we are locked into supporting overwrite_vars for the Dataset.merge
     # method due for backwards compatibility
     # TODO: consider deprecating it?
 
-    if isinstance(overwrite_vars, basestring):
+    if isinstance(overwrite_vars, str):
         overwrite_vars = set([overwrite_vars])
     overwrite_vars = set(overwrite_vars)
 
@@ -553,18 +585,32 @@ def dataset_merge_method(dataset, other, overwrite_vars, compat, join):
         objs = [dataset, other_no_overwrite, other_overwrite]
         priority_arg = 2
 
-    return merge_core(objs, compat, join, priority_arg=priority_arg)
+    return merge_core(objs, compat, join, priority_arg=priority_arg,
+                      fill_value=fill_value)
 
 
-def dataset_update_method(dataset, other):
+def dataset_update_method(
+    dataset: 'Dataset',
+    other: Union[
+        'Dataset',
+        Mapping[Hashable, Union[
+            'DataArray',
+            Variable,
+            Tuple[Hashable, Any],
+            Tuple[Sequence[Hashable], Any],
+        ]],
+    ],
+) -> Tuple['OrderedDict[Hashable, Variable]',
+           Set[Hashable],
+           Dict[Hashable, int]]:
     """Guts of the Dataset.update method.
 
     This drops a duplicated coordinates from `other` if `other` is not an
     `xarray.Dataset`, e.g., if it's a dict with DataArray values (GH2068,
     GH2180).
     """
+    from .dataarray import DataArray  # noqa: F811
     from .dataset import Dataset
-    from .dataarray import DataArray
 
     if not isinstance(other, Dataset):
         other = OrderedDict(other)
