@@ -1,37 +1,38 @@
-from __future__ import absolute_import, division, print_function
-
+import functools
 import warnings
+from collections import OrderedDict
 from distutils.version import LooseVersion
 
 import numpy as np
 
-from . import dtypes
+from . import dtypes, duck_array_ops, utils
 from .dask_array_ops import dask_rolling_wrapper
-from .ops import (
-    bn, has_bottleneck, inject_bottleneck_rolling_methods,
-    inject_datasetrolling_methods)
-from .pycompat import OrderedDict, dask_array_type, zip
+from .ops import inject_coarsen_methods
+from .pycompat import dask_array_type
+
+try:
+    import bottleneck
+except ImportError:
+    # use numpy methods instead
+    bottleneck = None
 
 
-def _get_new_dimname(dims, new_dim):
-    """ Get an new dimension name based on new_dim, that is not used in dims.
-    If the same name exists, we add an underscore(s) in the head.
+_ROLLING_REDUCE_DOCSTRING_TEMPLATE = """\
+Reduce this object's data windows by applying `{name}` along its dimension.
 
-    Example1:
-        dims: ['a', 'b', 'c']
-        new_dim: ['_rolling']
-        -> ['_rolling']
-    Example2:
-        dims: ['a', 'b', 'c', '_rolling']
-        new_dim: ['_rolling']
-        -> ['__rolling']
-    """
-    while new_dim in dims:
-        new_dim = '_' + new_dim
-    return new_dim
+Parameters
+----------
+**kwargs : dict
+    Additional keyword arguments passed on to `{name}`.
+
+Returns
+-------
+reduced : same type as caller
+    New object with `{name}` applied along its rolling dimnension.
+"""
 
 
-class Rolling(object):
+class Rolling:
     """A object that implements the moving window pattern.
 
     See Also
@@ -70,8 +71,8 @@ class Rolling(object):
         rolling : type of input argument
         """
 
-        if (has_bottleneck and
-                (LooseVersion(bn.__version__) < LooseVersion('1.0'))):
+        if (bottleneck is not None and
+                (LooseVersion(bottleneck.__version__) < LooseVersion('1.0'))):
             warnings.warn('xarray requires bottleneck version of 1.0 or '
                           'greater for rolling operations. Rolling '
                           'aggregation methods will use numpy instead'
@@ -113,6 +114,34 @@ class Rolling(object):
     def __len__(self):
         return self.obj.sizes[self.dim]
 
+    def _reduce_method(name):
+        array_agg_func = getattr(duck_array_ops, name)
+        bottleneck_move_func = getattr(bottleneck, 'move_' + name, None)
+
+        def method(self, **kwargs):
+            return self._numpy_or_bottleneck_reduce(
+                array_agg_func, bottleneck_move_func, **kwargs)
+        method.__name__ = name
+        method.__doc__ = _ROLLING_REDUCE_DOCSTRING_TEMPLATE.format(name=name)
+        return method
+
+    argmax = _reduce_method('argmax')
+    argmin = _reduce_method('argmin')
+    max = _reduce_method('max')
+    min = _reduce_method('min')
+    mean = _reduce_method('mean')
+    prod = _reduce_method('prod')
+    sum = _reduce_method('sum')
+    std = _reduce_method('std')
+    var = _reduce_method('var')
+    median = _reduce_method('median')
+
+    def count(self):
+        rolling_count = self._counts()
+        enough_periods = rolling_count >= self._min_periods
+        return rolling_count.where(enough_periods)
+    count.__doc__ = _ROLLING_REDUCE_DOCSTRING_TEMPLATE.format(name='count')
+
 
 class DataArrayRolling(Rolling):
     def __init__(self, obj, windows, min_periods=None, center=False):
@@ -149,7 +178,7 @@ class DataArrayRolling(Rolling):
         Dataset.rolling
         Dataset.groupby
         """
-        super(DataArrayRolling, self).__init__(
+        super().__init__(
             obj, windows, min_periods=min_periods, center=center)
 
         self.window_labels = self.obj[self.dim]
@@ -189,15 +218,15 @@ class DataArrayRolling(Rolling):
         --------
         >>> da = DataArray(np.arange(8).reshape(2, 4), dims=('a', 'b'))
         >>>
-        >>> rolling = da.rolling(a=3)
-        >>> rolling.to_datarray('window_dim')
+        >>> rolling = da.rolling(b=3)
+        >>> rolling.construct('window_dim')
         <xarray.DataArray (a: 2, b: 4, window_dim: 3)>
         array([[[np.nan, np.nan, 0], [np.nan, 0, 1], [0, 1, 2], [1, 2, 3]],
                [[np.nan, np.nan, 4], [np.nan, 4, 5], [4, 5, 6], [5, 6, 7]]])
         Dimensions without coordinates: a, b, window_dim
         >>>
-        >>> rolling = da.rolling(a=3, center=True)
-        >>> rolling.to_datarray('window_dim')
+        >>> rolling = da.rolling(b=3, center=True)
+        >>> rolling.construct('window_dim')
         <xarray.DataArray (a: 2, b: 4, window_dim: 3)>
         array([[[np.nan, 0, 1], [0, 1, 2], [1, 2, 3], [2, 3, np.nan]],
                [[np.nan, 4, 5], [4, 5, 6], [5, 6, 7], [6, 7, np.nan]]])
@@ -230,8 +259,31 @@ class DataArrayRolling(Rolling):
         -------
         reduced : DataArray
             Array with summarized data.
+
+        Examples
+        --------
+        >>> da = DataArray(np.arange(8).reshape(2, 4), dims=('a', 'b'))
+        >>>
+        >>> rolling = da.rolling(b=3)
+        >>> rolling.construct('window_dim')
+        <xarray.DataArray (a: 2, b: 4, window_dim: 3)>
+        array([[[np.nan, np.nan, 0], [np.nan, 0, 1], [0, 1, 2], [1, 2, 3]],
+               [[np.nan, np.nan, 4], [np.nan, 4, 5], [4, 5, 6], [5, 6, 7]]])
+        Dimensions without coordinates: a, b, window_dim
+        >>>
+        >>> rolling.reduce(np.sum)
+        <xarray.DataArray (a: 2, b: 4)>
+        array([[nan, nan,  3.,  6.],
+               [nan, nan, 15., 18.]])
+        Dimensions without coordinates: a, b
+        >>>
+        >>> rolling = da.rolling(b=3, min_periods=1)
+        >>> rolling.reduce(np.nansum)
+        <xarray.DataArray (a: 2, b: 4)>
+        array([[ 0.,  1.,  3.,  6.],
+               [ 4.,  9., 15., 18.]])
         """
-        rolling_dim = _get_new_dimname(self.obj.dims, '_rolling_dim')
+        rolling_dim = utils.get_temp_dimname(self.obj.dims, '_rolling_dim')
         windows = self.construct(rolling_dim)
         result = windows.reduce(func, dim=rolling_dim, **kwargs)
 
@@ -242,7 +294,7 @@ class DataArrayRolling(Rolling):
     def _counts(self):
         """ Number of non-nan entries in each rolling window. """
 
-        rolling_dim = _get_new_dimname(self.obj.dims, '_rolling_dim')
+        rolling_dim = utils.get_temp_dimname(self.obj.dims, '_rolling_dim')
         # We use False as the fill_value instead of np.nan, since boolean
         # array is faster to be reduced than object array.
         # The use of skipna==False is also faster since it does not need to
@@ -253,71 +305,65 @@ class DataArrayRolling(Rolling):
                   .sum(dim=rolling_dim, skipna=False))
         return counts
 
-    @classmethod
-    def _reduce_method(cls, func):
-        """
-        Methods to return a wrapped function for any function `func` for
-        numpy methods.
-        """
+    def _bottleneck_reduce(self, func, **kwargs):
+        from .dataarray import DataArray
 
-        def wrapped_func(self, **kwargs):
-            return self.reduce(func, **kwargs)
-        return wrapped_func
+        # bottleneck doesn't allow min_count to be 0, although it should
+        # work the same as if min_count = 1
+        if self.min_periods is not None and self.min_periods == 0:
+            min_count = 1
+        else:
+            min_count = self.min_periods
 
-    @classmethod
-    def _bottleneck_reduce(cls, func):
-        """
-        Methods to return a wrapped function for any function `func` for
-        bottoleneck method, except for `median`.
-        """
+        axis = self.obj.get_axis_num(self.dim)
 
-        def wrapped_func(self, **kwargs):
-            from .dataarray import DataArray
-
-            # bottleneck doesn't allow min_count to be 0, although it should
-            # work the same as if min_count = 1
-            if self.min_periods is not None and self.min_periods == 0:
-                min_count = 1
-            else:
-                min_count = self.min_periods
-
-            axis = self.obj.get_axis_num(self.dim)
-
-            padded = self.obj.variable
-            if self.center:
-                if (LooseVersion(np.__version__) < LooseVersion('1.13') and
-                        self.obj.dtype.kind == 'b'):
-                    # with numpy < 1.13 bottleneck cannot handle np.nan-Boolean
-                    # mixed array correctly. We cast boolean array to float.
-                    padded = padded.astype(float)
-
-                if isinstance(padded.data, dask_array_type):
-                    # Workaround to make the padded chunk size is larger than
-                    # self.window-1
-                    shift = - (self.window + 1) // 2
-                    offset = (self.window - 1) // 2
-                    valid = (slice(None), ) * axis + (
-                        slice(offset, offset + self.obj.shape[axis]), )
-                else:
-                    shift = (-self.window // 2) + 1
-                    valid = (slice(None), ) * axis + (slice(-shift, None), )
-                padded = padded.pad_with_fill_value(**{self.dim: (0, -shift)})
+        padded = self.obj.variable
+        if self.center:
+            if (LooseVersion(np.__version__) < LooseVersion('1.13') and
+                    self.obj.dtype.kind == 'b'):
+                # with numpy < 1.13 bottleneck cannot handle np.nan-Boolean
+                # mixed array correctly. We cast boolean array to float.
+                padded = padded.astype(float)
 
             if isinstance(padded.data, dask_array_type):
-                values = dask_rolling_wrapper(func, padded,
-                                              window=self.window,
-                                              min_count=min_count,
-                                              axis=axis)
+                # Workaround to make the padded chunk size is larger than
+                # self.window-1
+                shift = - (self.window + 1) // 2
+                offset = (self.window - 1) // 2
+                valid = (slice(None), ) * axis + (
+                    slice(offset, offset + self.obj.shape[axis]), )
             else:
-                values = func(padded.data, window=self.window,
-                              min_count=min_count, axis=axis)
+                shift = (-self.window // 2) + 1
+                valid = (slice(None), ) * axis + (slice(-shift, None), )
+            padded = padded.pad_with_fill_value({self.dim: (0, -shift)})
 
-            if self.center:
-                values = values[valid]
-            result = DataArray(values, self.obj.coords)
+        if isinstance(padded.data, dask_array_type):
+            raise AssertionError('should not be reachable')
+            values = dask_rolling_wrapper(func, padded.data,
+                                          window=self.window,
+                                          min_count=min_count,
+                                          axis=axis)
+        else:
+            values = func(padded.data, window=self.window,
+                          min_count=min_count, axis=axis)
 
-            return result
-        return wrapped_func
+        if self.center:
+            values = values[valid]
+        result = DataArray(values, self.obj.coords)
+
+        return result
+
+    def _numpy_or_bottleneck_reduce(
+        self, array_agg_func, bottleneck_move_func, **kwargs
+    ):
+        if (bottleneck_move_func is not None and
+                not isinstance(self.obj.data, dask_array_type)):
+            # TODO: renable bottleneck with dask after the issues
+            # underlying https://github.com/pydata/xarray/issues/2940 are
+            # fixed.
+            return self._bottleneck_reduce(bottleneck_move_func, **kwargs)
+        else:
+            return self.reduce(array_agg_func, **kwargs)
 
 
 class DatasetRolling(Rolling):
@@ -355,7 +401,7 @@ class DatasetRolling(Rolling):
         Dataset.groupby
         DataArray.groupby
         """
-        super(DatasetRolling, self).__init__(obj, windows, min_periods, center)
+        super().__init__(obj, windows, min_periods, center)
         if self.dim not in self.obj.dims:
             raise KeyError(self.dim)
         # Keep each Rolling object as an OrderedDict
@@ -366,6 +412,16 @@ class DatasetRolling(Rolling):
                 self.rollings[key] = DataArrayRolling(
                     da, windows, min_periods, center)
 
+    def _dataset_implementation(self, func, **kwargs):
+        from .dataset import Dataset
+        reduced = OrderedDict()
+        for key, da in self.obj.data_vars.items():
+            if self.dim in da.dims:
+                reduced[key] = func(self.rollings[key], **kwargs)
+            else:
+                reduced[key] = self.obj[key]
+        return Dataset(reduced, coords=self.obj.coords)
+
     def reduce(self, func, **kwargs):
         """Reduce the items in this group by applying `func` along some
         dimension(s).
@@ -384,43 +440,20 @@ class DatasetRolling(Rolling):
         reduced : DataArray
             Array with summarized data.
         """
-        from .dataset import Dataset
-        reduced = OrderedDict()
-        for key, da in self.obj.data_vars.items():
-            if self.dim in da.dims:
-                reduced[key] = self.rollings[key].reduce(func, **kwargs)
-            else:
-                reduced[key] = self.obj[key]
-        return Dataset(reduced, coords=self.obj.coords)
+        return self._dataset_implementation(
+            functools.partial(DataArrayRolling.reduce, func=func), **kwargs)
 
     def _counts(self):
-        from .dataset import Dataset
-        reduced = OrderedDict()
-        for key, da in self.obj.data_vars.items():
-            if self.dim in da.dims:
-                reduced[key] = self.rollings[key]._counts()
-            else:
-                reduced[key] = self.obj[key]
-        return Dataset(reduced, coords=self.obj.coords)
+        return self._dataset_implementation(DataArrayRolling._counts)
 
-    @classmethod
-    def _reduce_method(cls, func):
-        """
-        Return a wrapped function for injecting numpy and bottoleneck methods.
-        see ops.inject_datasetrolling_methods
-        """
-
-        def wrapped_func(self, **kwargs):
-            from .dataset import Dataset
-            reduced = OrderedDict()
-            for key, da in self.obj.data_vars.items():
-                if self.dim in da.dims:
-                    reduced[key] = getattr(self.rollings[key],
-                                           func.__name__)(**kwargs)
-                else:
-                    reduced[key] = self.obj[key]
-            return Dataset(reduced, coords=self.obj.coords)
-        return wrapped_func
+    def _numpy_or_bottleneck_reduce(
+        self, array_agg_func, bottleneck_move_func, **kwargs
+    ):
+        return self._dataset_implementation(
+            functools.partial(DataArrayRolling._numpy_or_bottleneck_reduce,
+                              array_agg_func=array_agg_func,
+                              bottleneck_move_func=bottleneck_move_func),
+            **kwargs)
 
     def construct(self, window_dim, stride=1, fill_value=dtypes.NA):
         """
@@ -454,5 +487,119 @@ class DatasetRolling(Rolling):
             **{self.dim: slice(None, None, stride)})
 
 
-inject_bottleneck_rolling_methods(DataArrayRolling)
-inject_datasetrolling_methods(DatasetRolling)
+class Coarsen:
+    """A object that implements the coarsen.
+
+    See Also
+    --------
+    Dataset.coarsen
+    DataArray.coarsen
+    """
+
+    _attributes = ['windows', 'side', 'trim_excess']
+
+    def __init__(self, obj, windows, boundary, side, coord_func):
+        """
+        Moving window object.
+
+        Parameters
+        ----------
+        obj : Dataset or DataArray
+            Object to window.
+        windows : A mapping from a dimension name to window size
+            dim : str
+                Name of the dimension to create the rolling iterator
+                along (e.g., `time`).
+            window : int
+                Size of the moving window.
+        boundary : 'exact' | 'trim' | 'pad'
+            If 'exact', a ValueError will be raised if dimension size is not a
+            multiple of window size. If 'trim', the excess indexes are trimed.
+            If 'pad', NA will be padded.
+        side : 'left' or 'right' or mapping from dimension to 'left' or 'right'
+        coord_func: mapping from coordinate name to func.
+
+        Returns
+        -------
+        coarsen
+        """
+        self.obj = obj
+        self.windows = windows
+        self.side = side
+        self.boundary = boundary
+
+        if not utils.is_dict_like(coord_func):
+            coord_func = {d: coord_func for d in self.obj.dims}
+        for c in self.obj.coords:
+            if c not in coord_func:
+                coord_func[c] = duck_array_ops.mean
+        self.coord_func = coord_func
+
+    def __repr__(self):
+        """provide a nice str repr of our coarsen object"""
+
+        attrs = ["{k}->{v}".format(k=k, v=getattr(self, k))
+                 for k in self._attributes
+                 if getattr(self, k, None) is not None]
+        return "{klass} [{attrs}]".format(klass=self.__class__.__name__,
+                                          attrs=','.join(attrs))
+
+
+class DataArrayCoarsen(Coarsen):
+    @classmethod
+    def _reduce_method(cls, func):
+        """
+        Return a wrapped function for injecting numpy methods.
+        see ops.inject_coarsen_methods
+        """
+        def wrapped_func(self, **kwargs):
+            from .dataarray import DataArray
+
+            reduced = self.obj.variable.coarsen(
+                self.windows, func, self.boundary, self.side)
+            coords = {}
+            for c, v in self.obj.coords.items():
+                if c == self.obj.name:
+                    coords[c] = reduced
+                else:
+                    if any(d in self.windows for d in v.dims):
+                        coords[c] = v.variable.coarsen(
+                            self.windows, self.coord_func[c],
+                            self.boundary, self.side)
+                    else:
+                        coords[c] = v
+            return DataArray(reduced, dims=self.obj.dims, coords=coords)
+
+        return wrapped_func
+
+
+class DatasetCoarsen(Coarsen):
+    @classmethod
+    def _reduce_method(cls, func):
+        """
+        Return a wrapped function for injecting numpy methods.
+        see ops.inject_coarsen_methods
+        """
+        def wrapped_func(self, **kwargs):
+            from .dataset import Dataset
+
+            reduced = OrderedDict()
+            for key, da in self.obj.data_vars.items():
+                reduced[key] = da.variable.coarsen(
+                    self.windows, func, self.boundary, self.side)
+
+            coords = {}
+            for c, v in self.obj.coords.items():
+                if any(d in self.windows for d in v.dims):
+                    coords[c] = v.variable.coarsen(
+                        self.windows, self.coord_func[c],
+                        self.boundary, self.side)
+                else:
+                    coords[c] = v.variable
+            return Dataset(reduced, coords=coords)
+
+        return wrapped_func
+
+
+inject_coarsen_methods(DataArrayCoarsen)
+inject_coarsen_methods(DatasetCoarsen)
