@@ -1,6 +1,6 @@
-from __future__ import absolute_import, division, print_function
-
 import pickle
+from collections import OrderedDict
+from contextlib import suppress
 from distutils.version import LooseVersion
 from textwrap import dedent
 
@@ -11,7 +11,6 @@ import pytest
 import xarray as xr
 import xarray.ufuncs as xu
 from xarray import DataArray, Dataset, Variable
-from xarray.core.pycompat import OrderedDict, suppress
 from xarray.tests import mock
 
 from . import (
@@ -23,7 +22,7 @@ da = pytest.importorskip('dask.array')
 dd = pytest.importorskip('dask.dataframe')
 
 
-class DaskTestCase(object):
+class DaskTestCase:
     def assertLazyAnd(self, expected, actual, test):
 
         with (dask.config.set(scheduler='single-threaded')
@@ -587,7 +586,7 @@ class TestDataArrayAndDataset(DaskTestCase):
         self.assertLazyAndIdentical(self.lazy_array, a)
 
 
-class TestToDaskDataFrame(object):
+class TestToDaskDataFrame:
 
     def test_to_dask_dataframe(self):
         # Test conversion of Datasets to dask DataFrames
@@ -843,3 +842,16 @@ def test_basic_compute():
             ds.compute()
             ds.foo.compute()
             ds.foo.variable.compute()
+
+
+@pytest.mark.skipif(LooseVersion(dask.__version__) < LooseVersion('0.20.0'),
+                    reason='needs newer dask')
+def test_dask_layers_and_dependencies():
+    ds = Dataset({'foo': ('x', range(5)),
+                  'bar': ('x', range(5))}).chunk()
+
+    x = dask.delayed(ds)
+    assert set(x.__dask_graph__().dependencies).issuperset(
+        ds.__dask_graph__().dependencies)
+    assert set(x.foo.__dask_graph__().dependencies).issuperset(
+        ds.__dask_graph__().dependencies)
