@@ -1,5 +1,3 @@
-from __future__ import absolute_import, division, print_function
-
 import itertools
 
 import numpy as np
@@ -8,14 +6,13 @@ import pytest
 
 from xarray import DataArray, Dataset, Variable
 from xarray.core import indexing, nputils
-from xarray.core.pycompat import native_int_types
 
 from . import IndexerMaker, ReturnItem, assert_array_equal, raises_regex
 
 B = IndexerMaker(indexing.BasicIndexer)
 
 
-class TestIndexers(object):
+class TestIndexers:
     def set_to_zero(self, x, i):
         x = x.copy()
         x[i] = 0
@@ -46,6 +43,22 @@ class TestIndexers(object):
         assert res[0] == (0,)
         assert res[1] == (1,)
 
+    def test_stacked_multiindex_min_max(self):
+        data = np.random.randn(3, 23, 4)
+        da = DataArray(
+            data, name="value",
+            dims=["replicate", "rsample", "exp"],
+            coords=dict(
+                replicate=[0, 1, 2],
+                exp=["a", "b", "c", "d"],
+                rsample=list(range(23))
+            ),
+        )
+        da2 = da.stack(sample=("replicate", "rsample"))
+        s = da2.sample
+        assert_array_equal(da2.loc['a', s.max()], data[2, 22, 0])
+        assert_array_equal(da2.loc['b', s.min()], data[0, 0, 1])
+
     def test_convert_label_indexer(self):
         # TODO: add tests that aren't just for edge cases
         index = pd.Index([1, 2, 3])
@@ -132,7 +145,7 @@ class TestIndexers(object):
                      pd.MultiIndex.from_product([[1, 2], [-1, -2]]))
 
 
-class TestLazyArray(object):
+class TestLazyArray:
     def test_slice_slice(self):
         I = ReturnItem()  # noqa: E741  # allow ambiguous name
         for size in [100, 99]:
@@ -174,7 +187,7 @@ class TestLazyArray(object):
                                           indexing.LazilyOuterIndexedArray)
 
                         # make sure actual.key is appropriate type
-                        if all(isinstance(k, native_int_types + (slice, ))
+                        if all(isinstance(k, (int, slice, ))
                                for k in v_lazy._data.key.tuple):
                             assert isinstance(v_lazy._data.key,
                                               indexing.BasicIndexer)
@@ -247,7 +260,7 @@ class TestLazyArray(object):
         check_indexing(v_eager, v_lazy, indexers)
 
 
-class TestCopyOnWriteArray(object):
+class TestCopyOnWriteArray:
     def test_setitem(self):
         original = np.arange(10)
         wrapped = indexing.CopyOnWriteArray(original)
@@ -271,7 +284,7 @@ class TestCopyOnWriteArray(object):
         assert np.array(x[B[0]][B[()]]) == 'foo'
 
 
-class TestMemoryCachedArray(object):
+class TestMemoryCachedArray:
     def test_wrapper(self):
         original = indexing.LazilyOuterIndexedArray(np.arange(10))
         wrapped = indexing.MemoryCachedArray(original)
@@ -340,7 +353,7 @@ def check_integer(indexer_cls):
 def check_slice(indexer_cls):
     (value,) = indexer_cls((slice(1, None, np.int64(2)),)).tuple
     assert value == slice(1, None, 2)
-    assert isinstance(value.step, native_int_types)
+    assert isinstance(value.step, int)
 
 
 def check_array1d(indexer_cls):
@@ -384,7 +397,7 @@ def test_vectorized_indexer():
                                     np.arange(5, dtype=np.int64)))
 
 
-class Test_vectorized_indexer(object):
+class Test_vectorized_indexer:
     @pytest.fixture(autouse=True)
     def setup(self):
         self.data = indexing.NumpyIndexingAdapter(np.random.randn(10, 12, 13))
@@ -508,13 +521,20 @@ def test_decompose_indexers(shape, indexer_mode, indexing_support):
 
 
 def test_implicit_indexing_adapter():
-    array = np.arange(10)
+    array = np.arange(10, dtype=np.int64)
     implicit = indexing.ImplicitToExplicitIndexingAdapter(
         indexing.NumpyIndexingAdapter(array), indexing.BasicIndexer)
     np.testing.assert_array_equal(array, np.asarray(implicit))
     np.testing.assert_array_equal(array, implicit[:])
 
 
+def test_implicit_indexing_adapter_copy_on_write():
+    array = np.arange(10, dtype=np.int64)
+    implicit = indexing.ImplicitToExplicitIndexingAdapter(
+        indexing.CopyOnWriteArray(array))
+    assert isinstance(implicit[:], indexing.ImplicitToExplicitIndexingAdapter)
+
+
 def test_outer_indexer_consistency_with_broadcast_indexes_vectorized():
     def nonzero(x):
         if isinstance(x, np.ndarray) and x.dtype.kind == 'b':
